// -*- c-basic-offset: 8; inent-tabs-mode: nil -*-
//
//  SqliteCommand.cs
//
//  Author(s): Vladimir Vukicevic  <vladimir@pobox.com>
//
//  Copyright (C) 2002  Vladimir Vukicevic
//
//  Modified 2003-07-25 Mars Tanumihardja <marst@vulcan.com>
//	to support running on MS .NET on Windows.
//

using System;
using System.Text;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Data;
using System.Diagnostics;

namespace Mono.Data.SqliteClient 
{
	public class SqliteCommand : IDbCommand
	{
		SqliteConnection parent_conn;
		//                SqliteTransaction transaction;
		IDbTransaction transaction;
		string sql;
		int timeout;
		CommandType type;
		UpdateRowSource upd_row_source;
		SqliteParameterCollection sql_params;

		public SqliteCommand ()
		{
			sql = "";
			sql_params = new SqliteParameterCollection ();
		}

		public SqliteCommand (string sqlText, SqliteConnection dbConn)
		{
			sql = sqlText;
			parent_conn = dbConn;
			sql_params = new SqliteParameterCollection ();
		}

		public SqliteCommand (string sqlText, SqliteConnection dbConn, IDbTransaction trans)
		{
			sql = sqlText;
			parent_conn = dbConn;
			transaction = trans;
			sql_params = new SqliteParameterCollection ();
		}

		public void Dispose ()
		{
		}

		public string CommandText 
		{
			get 
			{
				return sql;
			}
			set 
			{
				sql = value;
			}
		}

		// note that we could actually implement
		// a timeout with sqlite, but setting up a signal to interrupt us after
		// a certain amount of time, but it's probably not worth the effort
		public int CommandTimeout 
		{
			get 
			{
				return timeout;
			}
			set 
			{
				timeout = value;
			}
		}

		IDbConnection IDbCommand.Connection 
		{
			get 
			{
				return parent_conn;
			}
			set 
			{
				if (!(value is SqliteConnection)) 
				{
					throw new InvalidOperationException ("Can't set Connection to something other than a SqliteConnection");
				}
				parent_conn = (SqliteConnection) value;
			}
		}

		public SqliteConnection Connection 
		{
			get 
			{
				return parent_conn;
			}
			set 
			{
				parent_conn = value;
			}
		}

		public CommandType CommandType 
		{
			get 
			{
				return type;
			}
			set 
			{
				type = value;
			}
		}

		IDataParameterCollection IDbCommand.Parameters 
		{
			get 
			{
				return Parameters;
			}
		}

		public SqliteParameterCollection Parameters 
		{
			get 
			{
				return sql_params;
			}
		}

		public IDbTransaction Transaction 
		{
			get 
			{
				return transaction;
			}
			set 
			{
				transaction = value;
			}
		}

		public UpdateRowSource UpdatedRowSource 
		{
			get 
			{
				return upd_row_source;
			}
			set 
			{
				upd_row_source = value;
			}
		}

		public void Prepare ()
		{
		}

		public void Cancel ()
		{
		}

		IDbDataParameter IDbCommand.CreateParameter ()
		{
			return CreateParameter ();
		}

		public SqliteParameter CreateParameter ()
		{
			return new SqliteParameter ();
		}

		public int ExecuteNonQuery ()
		{
			int rows_affected;
			SqliteDataReader r = ExecuteReader (CommandBehavior.Default, false, out rows_affected);
			return rows_affected;
		}

		public object ExecuteScalar ()
		{
			SqliteDataReader r = ExecuteReader ();
			if (r == null || !r.Read ()) 
			{
				return null;
			}
			object o = r[0];
			r.Close ();
			return o;
		}

		IDataReader IDbCommand.ExecuteReader ()
		{
			return ExecuteReader ();
		}

		IDataReader IDbCommand.ExecuteReader (CommandBehavior behavior)
		{
			return ExecuteReader (behavior);
		}

		public SqliteDataReader ExecuteReader ()
		{
			return ExecuteReader (CommandBehavior.Default);
		}

		public SqliteDataReader ExecuteReader (CommandBehavior behavior)
		{
			int r;
			return ExecuteReader (behavior, true, out r);
		}

		public SqliteDataReader ExecuteReader (CommandBehavior behavior, bool want_results, out int rows_affected)
		{
			SqliteDataReader reader = null;
			SqliteError err;

			parent_conn.StartExec ();

			string msg = "";
			unsafe 
			{
				try 
				{
					if (want_results) 
					{
						reader = new SqliteDataReader (this);

						// MarsT: calling sqlite_exec to get result set back requires using a callback method.
						// Unfortunately, with .NET 1.0 and 1.1, PInvoke doesn't support marshaling delegates  
						// for the CollingConvention.Cdecl; it only supports StdCall (WinApi).
						// See ".NET and COM" by Adam Nathan Chapter 19, page 815 (Caution note).

//						SqliteCallbackFunction callback = new SqliteCallbackFunction (reader.SqliteCallback);
//
//						// NullReferenceException is throw here on return... why?
//						err = sqlite_exec (
//							parent_conn.Handle,
//							sql,
//							callback,
//							IntPtr.Zero, 
//							//&msg_result
//							out msg
//							);

						// Here we make use of sqlite's alternative access interface that does not use 
						// callbacks. (See http://www.sqlite.org/c_interface.html)

						IntPtr virtualMachine = IntPtr.Zero;
						string sqltail = null;

						err = sqlite_compile(parent_conn.Handle, sql, out sqltail, &virtualMachine, out msg);

						int colNum = 0;
						sbyte** colData;
						sbyte** colNames;
						while (true) 
						{
							err = sqlite_step(virtualMachine, out colNum, out colData, out colNames);
							if (err == SqliteError.Row) 
								reader.SqliteCallback(IntPtr.Zero, colNum, colData, colNames);
							else
								break;
						}

						err = sqlite_finalize(virtualMachine, out msg);

						reader.ReadingDone ();
					} 
					else 
					{
						err = sqlite_exec (parent_conn.Handle, sql, null, IntPtr.Zero, out msg);
					}
				} 
				finally 
				{
					parent_conn.EndExec ();
				}
			}

			if (err != SqliteError.OK)
				throw new ApplicationException ("Sqlite error " + msg);

			rows_affected = NumChanges ();
			return reader;
		}

		internal int NumChanges () 
		{
			return sqlite_changes (parent_conn.Handle);
		}

		internal unsafe delegate int SqliteCallbackFunction (IntPtr pArg, int argc, sbyte** argv,  sbyte** colnames);

		[DllImport("sqlite", CallingConvention=CallingConvention.Cdecl)]
		unsafe static extern SqliteError sqlite_exec (IntPtr handle, string sql, SqliteCallbackFunction callback,
			IntPtr user_data, out string errstr_ptr);

		[DllImport("sqlite", CallingConvention=CallingConvention.Cdecl)]
		static extern int sqlite_changes (IntPtr handle);


		[DllImport("sqlite", CallingConvention=CallingConvention.Cdecl)]
		static unsafe extern SqliteError sqlite_compile (IntPtr handle, string sql, out string sqltail, IntPtr* vm, out string errstr_ptr);

		[DllImport("sqlite", CallingConvention=CallingConvention.Cdecl)]
		static unsafe extern SqliteError sqlite_step (IntPtr vm, out int colnum, out sbyte** coldata, out sbyte** colnames);

		[DllImport("sqlite", CallingConvention=CallingConvention.Cdecl)]
		static extern SqliteError sqlite_finalize (IntPtr vm, out string errstr_ptr);


		internal enum SqliteError : int 
		{
			OK,
			Error,
			Internal,
			Perm,
			Abort,
			Busy,
			Locked,
			NoMem,
			ReadOnly,
			Interrupt,
			IOErr,
			Corrupt,
			NotFound,
			Full,
			CantOpen,
			Protocol,
			Empty,
			Schema,
			TooBig,
			Constraint,
			Mismatch,
			Misuse,
			NoLfs,
			Auth,
			Row = 100,
			Done = 101
		}
	}
}
